/*
    This file is part of task -- a task scheduler for distributed computations
    Copyright (C) 2012  Laurent Poirrier

    task is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <poll.h>
#include <stdio.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "network.h"
#include "block.h"
#include "protocol.h"
#include "common.h"
#include "config.h"
#include "buffer.h"
#include "control.h"

// **************************************************************************
// 
// **************************************************************************
#define DEBUG(s...)	fprintf(stderr, s)

#define CONTROL_SEND_ONLY	1

typedef struct control_t {
	const char *action;
	int fd;
	int flags;
	buffer inbuff;
	buffer outbuff;
} control;

// **************************************************************************
// 
// **************************************************************************
static int control_init(control *k, const char *action,
	const char *node, const char *service)
{
	k->action = action;

	k->fd = netcli_init(node, service);
	
	if (k->fd < 0)
		return(-1);
	
	k->flags = 0;
	buffer_init(&k->inbuff, 0);
	buffer_init(&k->outbuff, 0);
	return(0);
}

static void control_clear(control *k)
{
	close(k->fd);
	buffer_clear(&k->inbuff);
	buffer_clear(&k->outbuff);
}


// **************************************************************************
// 
// **************************************************************************
// Warning: sdump() overshoots by one, be sure that is ok
void sdump(int h, int tid, int cid, char *host, char *d, size_t sz)
{
	FILE *f;
	size_t o, z;
	
	f = (h == PROTO_SV_OUT) ? stdout : stderr;
	
	o = 0;
	while (o < sz) {
		z = o;
		while ((z < sz) && (d[z] != '\n'))
			z++;
		
		d[z] = 0;
		
		fprintf(f, "[%d](%d %s): %s%s\n",
			tid, cid, host, d + o, (z < sz) ? "" : "[...]");

		o = z + 1;
	}
}

// **************************************************************************
// 
// **************************************************************************
static int control_iter(control *k)
{
	void *d;
	size_t dsz;
	int tid, cid;
	char *hs, *s;
	int h;
	
	if (proto_prepare(&k->inbuff, PROTO_SV_MAGIC, conf.proto_buff_max)) {
		DEBUG("%s: protocol error\n", k->action);
		return(-1);
	}
	
	if (!proto_ready(&k->inbuff))
		return(0);
	
	h = proto_header(&k->inbuff);

	// ************************************	
	if (h == PROTO_SV_MSG) {
		if (proto_fetch(&k->inbuff, "d", &d, &dsz)) {
			DEBUG("%s: PROTO_SV_MSG: bad packet\n", k->action);
			return(-1);
		}

		if (block_write(1, d, dsz))
			return(-1);

		return(0);
	}

	// ************************************	
	if ((h == PROTO_SV_OUT) || (h == PROTO_SV_ERR)) {
		hs = (h == PROTO_SV_OUT) ? "OUT" : "ERR";
		
		if (proto_fetch(&k->inbuff, "iisd", &tid, &cid, &s, &d, &dsz)) {
			DEBUG("%s: PROTO_SV_%s: bad packet\n", k->action, hs);
			return(-1);
		}
		
		// sdump() overshoots by one, ensure that's ok
		buffer_alloc(&k->inbuff, k->inbuff.size + 1);
		sdump(h, tid, cid, s, (char *)d, dsz);
		
		return(0);
	}

	// ************************************	
	DEBUG("%s: unknown response header 0x%08x\n",
		k->action, h);
	return(-1);
}

// **************************************************************************
// 
// **************************************************************************
static int control_loop(control *k)
{
	struct pollfd pfd;
	int done, error;
	
	done = 0;
	error = 0;
	while (1) {
		pfd.fd = k->fd;
		pfd.events = POLLIN;
		if (!buffer_io_complete(&k->outbuff))
			pfd.events |= POLLOUT;
		pfd.revents = 0;
	
		if (poll(&pfd, 1, -1) < 0) {
			fprintf(stderr, "%s: poll(): %s\n",
				k->action, strerror(errno));
			return(1);
		}
		
		if (pfd.revents & POLLIN) {
			buffer_io_read(&k->inbuff, k->fd);
			done |= k->inbuff.eof;
			error |= k->inbuff.error;
		}
		
		if (pfd.revents & POLLOUT) {
			buffer_io_write(&k->outbuff, k->fd);
			// This flag is not used anymore, but still works
			if (k->flags & CONTROL_SEND_ONLY)
				done |= buffer_io_complete(&k->outbuff);
			error |= k->outbuff.eof | k->outbuff.error;
		}
		
		if ((done) || (error)) {
			error |= (k->inbuff.partial != 0);
			error |= !buffer_io_complete(&k->outbuff);

			control_clear(k);
			
			if (error) {
				fprintf(stderr, "%s: i/o error\n", k->action);
				return(1);
			} else {
				// fprintf(stderr, "-- ok --\n");
				return(0);
			}
		}
		
		if (control_iter(k))
			return(1);
	}
}

// **************************************************************************
// 
// **************************************************************************
int task_add(const char *node, const char *service,
	const char *dir, const char *cmd, int n, char **names)
{
	control k;
	buffer tcmd;
	int i;
	
	if (control_init(&k, "task add", node, service))
		return(1);
	
	if (!dir)
		dir = "";
	
	if (!cmd)
		cmd = "";
	
	buffer_init(&tcmd, 0);
	
	for (i = 0; i < n; i++) {
		buffer_reset(&tcmd);
		buffer_fappend(&tcmd, "task=\"%s\"\n%s", names[i], cmd);
		
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_ADD,
			"sss", dir, buffer_str(&tcmd), names[i]);
	}
	
	buffer_clear(&tcmd);
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");
	
	return(control_loop(&k));
}

// **************************************************************************
// 
// **************************************************************************
int task_rm(const char *node, const char *service,
	int flags, const char *dir, int n, char **names)
{
	control k;
	char *defz;
	int i;
	
	if (control_init(&k, "task stop", node, service))
		return(1);
	
	if (!dir)
		dir = "";

	if (n == 0) {
		defz = "";
		names = &defz;
		n = 1;
	}
	
	for (i = 0; i < n; i++) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_RM,
			"iss", flags, dir, names[i]);
	}
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");

	return(control_loop(&k));
}

// **************************************************************************
// 
// **************************************************************************
int task_exec(const char *node, const char *service,
	const char *cmd)
{
	control k;
	
	if (control_init(&k, "task exec", node, service))
		return(1);
	
	if (!cmd)
		cmd = "";
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_EXEC, "s", cmd);
	
	return(control_loop(&k));
}

// **************************************************************************
// 
// **************************************************************************
int task_list(const char *node, const char *service, int flags)
{
	control k;
	
	if (control_init(&k, "task list", node, service))
		return(1);
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_LIST, "i", flags);
	
	return(control_loop(&k));
}

// **************************************************************************
// 
// **************************************************************************
int task_nodes(const char *node, const char *service)
{
	control k;
	
	if (control_init(&k, "task nodes", node, service))
		return(1);
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_NODES, "");
	
	return(control_loop(&k));
}


