/*
    This file is part of task -- a task scheduler for distributed computations
    Copyright (C) 2012  Laurent Poirrier

    task is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "config.h"
#include "strlist.h"
#include "server.h"
#include "client.h"
#include "control.h"
#include "param.h"
#include "common.h"
#include "main.h"

// ****************************************************************************
// Server
// ****************************************************************************
static int main_server(int n, char **a)
{
	int r, flags;
	char *port;
	param p[4] = {
		{ PARAM_STRING, 0, -1, "-h", "host", NULL, NULL, 0, 1 },
		{ PARAM_STRING, 0, -1, "-p", "port", NULL, NULL, 0, 1 },
		{ PARAM_SWITCH, 0, -1, "-A", "", NULL, NULL, 0, 1 },
		{ PARAM_SWITCH, 0, -1, "-L", "", NULL, NULL, 0, 1 },
	};
	param_set ps = { "task server", 4, p };
	
	if (n < 0) {
		param_help(stdout, &ps);
		return(0);
	}
	
	if (param_init(&ps, n, a))
		return(1);
	
	if (p[2].value.b)
		flags = NETSERV_ALL;
	else if (p[3].value.b)
		flags = NETSERV_LOOPBACK;
	else if (p[0].value.s)
		flags = 0;
	else
		flags = NETSERV_LOOPBACK;
	
	port = (p[1].value.s) ? p[1].value.s : conf.default_service;
	
	r = task_server(p[0].value.s, port, flags);
	
	param_clear(&ps);
	
	return(r);
}

// ****************************************************************************
// Client
// ****************************************************************************
static int main_client(int n, char **a)
{
	int r;
	char *port;
	param p[4] = {
		{ PARAM_STRING, 0, -1, "-h", "host", NULL, NULL, 0, 1 },
		{ PARAM_STRING, 0, -1, "-p", "port", NULL, NULL, 0, 1 },
		{ PARAM_INT,    0, -1, "-n", "processes", NULL, NULL, 0, 1 },
		{ PARAM_STRING, 0, -1, "-wc", "command", "-ws", "file", 0, 1 },
	};
	param_set ps = { "task client", 4, p };

	if (n < 0) {
		param_help(stdout, &ps);
		return(0);
	}
	
	if (param_init(&ps, n, a))
		return(1);
	
	port = (p[1].value.s) ? p[1].value.s : conf.default_service;

	r = task_client(p[0].value.s, port, p[2].value.i, p[3].value.s);
	
	param_clear(&ps);
	return(r);
}

// ****************************************************************************
// Add
// ****************************************************************************
static int main_add(int n, char **a)
{
	int r;
	char *port;
	param p[5] = {
		{ PARAM_STRING,  0, -1, "-h", "host", NULL, NULL, 0, 1 },
		{ PARAM_STRING,  0, -1, "-p", "port", NULL, NULL, 0, 1 },
		{ PARAM_STRING,  0, -1, "-o", "dir", NULL, NULL, 0, 1 },
		{ PARAM_STRING,  0, -1, "-c", "command", "-s", "file", 1, 1 },
		{ PARAM_STRLIST, 0, -1, "", "name", "-a", "file", 1, 1 },
	};
	param_set ps = { "task add", 5, p };

	if (n <= 0) {
		param_help(stdout, &ps);
		return(0);
	}
	
	if (param_init(&ps, n, a))
		return(1);
	
	port = (p[1].value.s) ? p[1].value.s : conf.default_service;

	r = task_add(p[0].value.s, port, p[2].value.s, p[3].value.s,
		p[4].value.l.n, strlist_data(&p[4].value.l));
	
	param_clear(&ps);
	return(r);
}

// ****************************************************************************
// Remove
// ****************************************************************************
static int main_rm(int n, char **a)
{
	int r, flags;
	char *port;
	char *s;
	param p[5] = {
		{ PARAM_STRING,  0, -1, "-h", "host", NULL, NULL, 0, 1 },
		{ PARAM_STRING,  0, -1, "-p", "port", NULL, NULL, 0, 1 },
		{ PARAM_STRING,  0, -1, "-o", "dir", NULL, NULL, 0, 1 },
		{ PARAM_STRING,  0, -1, "-f", "d|r|p...", NULL, NULL, 0, 1 },
		{ PARAM_STRLIST, 0, -1, "", "name", "-a", "file", 0, 1 },
	};
	param_set ps = { "task rm", 5, p };
	
	if (n <= 0) {
		// 'task rm' without argument is actually valid and would
		// remove all tasks. But as it is somewhat dangerous, we
		// display the help in that case. Use 'task rm -f drp'
		// instead.
		param_help(stdout, &ps);
		return(0);
	}
	
	if (param_init(&ps, n, a))
		return(1);
		
	port = (p[1].value.s) ? p[1].value.s : conf.default_service;

	flags = 0;	
	if (p[3].value.s) {
		s = p[3].value.s;
		
		while (*s != 0) {
			if (*s == 'd')
				flags |= PROTO_TASK_DONE;
			else if (*s == 'r')
				flags |= PROTO_TASK_RUNNING;
			else if (*s == 'p')
				flags |= PROTO_TASK_PENDING;
			s++;
		}
	}

	if (flags == 0)
		flags = PROTO_TASK_DONE
			| PROTO_TASK_RUNNING | PROTO_TASK_PENDING;

	r = task_rm(p[0].value.s, port, flags, p[2].value.s,
		p[4].value.l.n, strlist_data(&p[4].value.l));

	param_clear(&ps);
	return(r);
}

// ****************************************************************************
// Exec
// ****************************************************************************
static int main_exec(int n, char **a)
{
	int r;
	char *port;
	param p[3] = {
		{ PARAM_STRING, 0, -1, "-h", "host", NULL, NULL, 0, 1 },
		{ PARAM_STRING, 0, -1, "-p", "port", NULL, NULL, 0, 1 },
		{ PARAM_STRING, 0, -1, "-c", "command", "-s", "file", 1, 1 },
	};
	param_set ps = { "task exec", 3, p };
	
	if (n <= 0) {
		param_help(stdout, &ps);
		return(0);
	}
	
	if (param_init(&ps, n, a))
		return(1);
	
	port = (p[1].value.s) ? p[1].value.s : conf.default_service;

	r = task_exec(p[0].value.s, port, p[2].value.s);
	
	param_clear(&ps);
	return(r);
}

// ****************************************************************************
// List tasks
// ****************************************************************************
static int main_list(int n, char **a)
{
	int r, flags, oflags;
	char *port;
	char *s;
	param p[8] = {
		{ PARAM_STRING, 0, -1, "-h", "host", NULL, NULL, 0, 1 },
		{ PARAM_STRING, 0, -1, "-p", "port", NULL, NULL, 0, 1 },
		{ PARAM_SWITCH, 0, -1, "-t", "", NULL, NULL, 0, 1 },
		{ PARAM_SWITCH, 0, -1, "-l", "", NULL, NULL, 0, 1 },
		{ PARAM_SWITCH, 0, -1, "done", "", NULL, NULL, 0, 1 },
		{ PARAM_SWITCH, 0, -1, "running", "", NULL, NULL, 0, 1 },
		{ PARAM_SWITCH, 0, -1, "pending", "", NULL, NULL, 0, 1 },
		{ PARAM_STRING, 0, -1, "", "d|r|p...", NULL, NULL, 0, 1 },
	};
	param_set ps = { "task list", 8, p };
	
	if (n < 0) {
		param_help(stdout, &ps);
		return(0);
	}
	
	if (param_init(&ps, n, a))
		return(1);
	
	port = (p[1].value.s) ? p[1].value.s : conf.default_service;

	flags = 0;
	oflags = 0;
	
	if (p[2].value.b)
		oflags |= PROTO_TASK_DURATIONS;
	if (p[3].value.b)
		oflags |= PROTO_TASK_TIMES;
	
	if (p[4].value.b)
		flags |= PROTO_TASK_DONE;
	if (p[5].value.b)
		flags |= PROTO_TASK_RUNNING;
	if (p[6].value.b)
		flags |= PROTO_TASK_PENDING;
	
	if (p[7].value.s) {
		s = p[7].value.s;
		
		while (*s != 0) {
			if (*s == 'd')
				flags |= PROTO_TASK_DONE;
			else if (*s == 'r')
				flags |= PROTO_TASK_RUNNING;
			else if (*s == 'p')
				flags |= PROTO_TASK_PENDING;
			s++;
		}
	}
	
	if (flags == 0)
		flags = PROTO_TASK_DONE | PROTO_TASK_RUNNING
			| PROTO_TASK_PENDING;
	
	r = task_list(p[0].value.s, port, flags | oflags);

	param_clear(&ps);
	return(r);
}

// ****************************************************************************
// Nodes info
// ****************************************************************************
static int main_nodes(int n, char **a)
{
	int r;
	char *port;
	param p[2] = {
		{ PARAM_STRING, 0, -1, "-h", "host", NULL, NULL, 0, 1 },
		{ PARAM_STRING, 0, -1, "-p", "port", NULL, NULL, 0, 1 },
	};
	param_set ps = { "task nodes", 2, p };
	
	if (n < 0) {
		param_help(stdout, &ps);
		return(0);
	}
	
	if (param_init(&ps, n, a))
		return(1);
		
	port = (p[1].value.s) ? p[1].value.s : conf.default_service;

	r = task_nodes(p[0].value.s, port);

	param_clear(&ps);
	return(r);
}

// ****************************************************************************
// Help
// ****************************************************************************
static int main_help(int n, char **a)
{
	printf("task v%d.%d -- Laurent Poirrier, 2012\n",
		TASK_VERSION, TASK_REVISION);
	
	//printf("\nConfig: shell: %s, host: %s, max file size: %zd\n",
	//	conf.shell, conf.host, conf.file_size_max);

	printf("\nServer:\n\n  ");
	main_server(-1, NULL);
	printf("\nClient:\n\n  ");
	main_client(-1, NULL);
	printf("\nControl:\n\n  ");
	main_add(-1, NULL);
	printf("  ");
	main_rm(-1, NULL);
	printf("  ");
	main_exec(-1, NULL);
	printf("  ");
	main_list(-1, NULL);
	printf("  ");
	main_nodes(-1, NULL);
	printf("\n");

	return(0);
}

// ****************************************************************************
// Main
// ****************************************************************************
int main(int argc, char **argv, char **envp)
{
	int r;
	
	config_init(envp);
	
	if (argc < 2)
		r = main_help(0, NULL);
	else if (!strcmp(argv[1], "help"))
		r = main_help(argc - 2, &argv[2]);
	else if (!strcmp(argv[1], "server"))
		r = main_server(argc - 2, &argv[2]);
	else if (!strcmp(argv[1], "client"))
		r = main_client(argc - 2, &argv[2]);
	else if (!strcmp(argv[1], "add"))
		r = main_add(argc - 2, &argv[2]);
	else if (!strcmp(argv[1], "exec"))
		r = main_exec(argc - 2, &argv[2]);
	else if (!strcmp(argv[1], "rm"))
		r = main_rm(argc - 2, &argv[2]);
	else if (!strcmp(argv[1], "list"))
		r = main_list(argc - 2, &argv[2]);
	else if (!strcmp(argv[1], "nodes"))
		r = main_nodes(argc - 2, &argv[2]);
	else {
		printf("Unknown subcommand: 'task %s'. Try 'task help'.\n",
			argv[1]);
		r = 1;
	}
	
	config_clear();
	
	return(r);
}

