/*
    This file is part of task -- a task scheduler for distributed computations
    Copyright (C) 2012  Laurent Poirrier

    task is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#define _POSIX_SOURCE
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include "network.h"


#define NETSERV_BACKLOG	10

static char nethost[1025];
static char netserv[32];

// **************************************************************************
// 
// **************************************************************************
int  netserv_init(const char *node, const char *service, int flags)
{
	struct addrinfo hints;
	struct addrinfo *res, *rp;
	int fd, e;
	
	if (flags & (NETSERV_ALL | NETSERV_LOOPBACK))
		node = NULL;
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;
	hints.ai_flags =
		((flags & NETSERV_ALL) ? AI_PASSIVE : 0) | AI_ADDRCONFIG;
	
	e = getaddrinfo(node, service, &hints, &res);
	
	if (e != 0) {
		fprintf(stderr, "netserv_init(): getaddrinfo(): %s\n",
			gai_strerror(e));
		if (e == EAI_SYSTEM)
			fprintf(stderr, "\t(%s)\n", strerror(errno));
		return(-1);
	}

	for (rp = res; rp != NULL; rp = rp->ai_next) {
		if (getnameinfo(rp->ai_addr, rp->ai_addrlen,
			nethost, sizeof(nethost), netserv, sizeof(netserv),
			NI_NUMERICHOST | NI_NUMERICSERV) == 0)
		{
			fprintf(stderr, "network: service %s resolved to "
				"%s %s\n", service, nethost, netserv);
		}
		
		fd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
		
		if (fd == -1) {
			fprintf(stderr, "netserv_init(): socket(): %s\n",
				strerror(errno));
			continue;
		}
	
		if (bind(fd, rp->ai_addr, rp->ai_addrlen)) {
			fprintf(stderr, "netserv_init(): bind(): %s\n",
				strerror(errno));
			close(fd);
			continue;
		}
		
		if (listen(fd, NETSERV_BACKLOG)) {
			fprintf(stderr, "netserv_init(): listen(): %s\n",
				strerror(errno));
			close(fd);
			continue;
		}
		
		if (fcntl(fd, F_SETFL, O_NONBLOCK)) {
			fprintf(stderr, "netserv_init(): "
				"fcntl(F_SETFL, O_NONBLOCK): %s\n",
				strerror(errno));
			close(fd);
			continue;
		}
		
		break;
	}
	
	freeaddrinfo(res);
	
	if (!rp)
		return(-1);
	
	return(fd);
}

// **************************************************************************
// 
// **************************************************************************
int netserv_accept(int sfd, netpeer *peer)
{
	struct sockaddr_storage pa;
	socklen_t pa_len;
	int cfd, e;
	
	pa_len = sizeof(pa);
	cfd = accept(sfd, (struct sockaddr *)&pa, &pa_len);
	
	if (cfd == -1) {
		fprintf(stderr, "netserv_accept(): accept(): %s\n",
			strerror(errno));
		return(-1);
	}
	
	if (fcntl(cfd, F_SETFL, O_NONBLOCK)) {
		fprintf(stderr, "netserv_accept(): "
			"fcntl(F_SETFL, O_NONBLOCK): %s\n",
			strerror(errno));
		close(cfd);
		return(-1);
	}

	if (peer) {
		e = getnameinfo((struct sockaddr *)&pa, pa_len,
			peer->host, sizeof(peer->host),
			peer->serv, sizeof(peer->serv), NI_NUMERICSERV);
		
		if (e) {
			fprintf(stderr, "netserv_accept(): getnameinfo(): %s\n",
				gai_strerror(e));
			if (e == EAI_SYSTEM)
				fprintf(stderr, "\t(%s)\n", strerror(errno));
			close(cfd);
			return(-1);
		}

		peer->fd = cfd;
	}
	
	return(cfd);
}



// **************************************************************************
// 
// **************************************************************************
int netcli_init(const char *node, const char *service)
{
	struct addrinfo hints;
	struct addrinfo *res, *rp;
	int fd, e;
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;
	hints.ai_flags = AI_ADDRCONFIG;
	
	e = getaddrinfo(node, service, &hints, &res);
	
	if (e != 0) {
		fprintf(stderr, "netserv_init(): getaddrinfo(): %s\n",
			gai_strerror(e));
		if (e == EAI_SYSTEM)
			fprintf(stderr, "\t(%s)\n", strerror(errno));
		return(-1);
	}

	for (rp = res; rp != NULL; rp = rp->ai_next) {
		e = getnameinfo(rp->ai_addr, rp->ai_addrlen,
			nethost, sizeof(nethost), netserv, sizeof(netserv),
			NI_NUMERICHOST | NI_NUMERICSERV);
		if (e) {
			fprintf(stderr, "netserv_init(): getnameinfo(): %s\n",
				gai_strerror(e));
			if (e == EAI_SYSTEM)
				fprintf(stderr, "\t(%s)\n", strerror(errno));
			nethost[0] = '?';
			nethost[1] = 0;
			netserv[0] = '?';
			nethost[1] = 0;
		}

		//fprintf(stderr, "network: %s %s resolved to %s %s\n",
		//	(node) ? node : "localhost",
		//	service, nethost, netserv);
	
		fd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
		
		if (fd == -1) {
			fprintf(stderr, "netserv_init(): socket(%s %s): %s\n",
				nethost, netserv, strerror(errno));
			continue;
		}
	
		if (connect(fd, rp->ai_addr, rp->ai_addrlen)) {
			fprintf(stderr, "netserv_init(): connect(%s %s): %s\n",
				nethost, netserv, strerror(errno));
			close(fd);
			continue;
		}
		
		if (fcntl(fd, F_SETFL, O_NONBLOCK)) {
			fprintf(stderr, "netcli_init(): "
				"fcntl(F_SETFL, O_NONBLOCK): %s\n",
				strerror(errno));
			close(fd);
			continue;
		}
		
		break;
	}
	
	freeaddrinfo(res);
	
	if (!rp)
		return(-1);
	
	return(fd);
}

// **************************************************************************
// 
// **************************************************************************


