/*
    This file is part of task -- a task scheduler for distributed computations
    Copyright (C) 2012  Laurent Poirrier

    task is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include "xmem.h"
#include "queue.h"


// ***************************************************************************
// 
// ***************************************************************************
task *task_new(int id, char *dir, char *cmd, char *name)
{
	task *t;
	
	t = xmalloc(sizeof(task));
	
	t->id = id;
	t->dir = xstrdup(dir);
	t->cmd = xstrdup(cmd);
	t->name = xstrdup(name);
	t->restrict_box = -1;
	t->forward_box = -1;
	t->outpath = xprintf(NULL, "%s%s%s.out",
		dir, ((dir[0]) ? "/" : ""), name);
	t->errpath = xprintf(NULL, "%s%s%s.log",
		dir, ((dir[0]) ? "/" : ""), name);
	t->outfd = -1;
	t->errfd = -1;
	t->start_time = 0;
	t->stop_time = 0;
	
	return(t);
}

// ***************************************************************************
// 
// ***************************************************************************
void task_free(task *t)
{
	xfree(t->dir);
	xfree(t->cmd);
	xfree(t->name);
	xfree(t->outpath);
	xfree(t->errpath);
	xfree(t);
}

// ***************************************************************************
// 
// ***************************************************************************
int task_open(task *t)
{
	if (t->dir[0]) {
		mkdir(t->dir, 0700);
	}
	
	t->outfd = creat(t->outpath, 0600);
	if (t->outfd < 0) {
		fprintf(stderr, "%s: %s (creating file)\n",
			t->outpath, strerror(errno));
		task_close(t, 1);
		return(-1);
	}
	
	t->errfd = creat(t->errpath, 0600);
	if (t->errfd < 0) {
		fprintf(stderr, "%s: %s (creating file)\n",
			t->errpath, strerror(errno));
		task_close(t, 1);
		return(-1);
	}
	
	t->start_time = time(NULL);
	t->stop_time = 0;
	
	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
void task_close(task *t, int rmfiles)
{
	if (t->outfd >= 0)
		close(t->outfd);
	if (t->errfd >= 0)
		close(t->errfd);
	
	if ((rmfiles) && (t->outfd >= 0)) {
		if (unlink(t->outpath)) {
			fprintf(stderr, "%s: %s (removing file)\n",
				t->outpath, strerror(errno));
		}
	}
	
	if ((rmfiles) && (t->errfd >= 0)) {
		if (unlink(t->errpath)) {
			fprintf(stderr, "%s: %s (removing file)\n",
				t->errpath, strerror(errno));
		}
	}
	
	t->outfd = -1;
	t->errfd = -1;
	
	t->stop_time = time(NULL);
}

// ***************************************************************************
// 
// ***************************************************************************
int task_matches(task *t, char *dir, char *name)
{
	return(	((!dir) || (!dir[0]) || (!strcmp(dir, t->dir)))
	&&	((!name) || (!name[0]) || (!strcmp(name, t->name))) );
}


// ***************************************************************************
// 
// ***************************************************************************
void queue_init(queue *q)
{
	q->head = NULL;
	q->tail = NULL;
	q->size = 0;
}

void queue_clear(queue *q)
{
	task *t, *u;
	
	t = q->head;
	
	while (t) {
		u = t->next;
		
		task_close(t, 0);
		task_free(t);
		
		t = u;
	}
	
	q->head = NULL;
	q->tail = NULL;
	q->size = 0;
}

// ***************************************************************************
// 
// ***************************************************************************
void queue_prepend(queue *q, task *t)
{
	t->prev = NULL;
	t->next = q->head;
	
	if (q->head)
		q->head->prev = t;
	
	q->head = t;
	
	if (!q->tail)
		q->tail = t;
	
	q->size++;
}

void queue_append(queue *q, task *t)
{
	t->prev = q->tail;
	t->next = NULL;
	
	if (q->tail)
		q->tail->next = t;
	
	q->tail = t;
	
	if (!q->head)
		q->head = t;
	
	q->size++;
}

void queue_unlink(queue *q, task *t)
{
	if (t->prev)
		t->prev->next = t->next;
	
	if (t->next)
		t->next->prev = t->prev;

	if (q->head == t)
		q->head = t->next;
	
	if (q->tail == t)
		q->tail = t->prev;
	
	t->prev = NULL;
	t->next = NULL;
	
	q->size--;
}

// ***************************************************************************
// 
// ***************************************************************************
task *queue_find(queue *q, int id)
{
	task *t;
	
	t = q->head;
	
	while (t) {
		if (t->id == id)
			return(t);
		
		t = t->next;
	}
	
	return(NULL);
}


