/*
    This file is part of task -- a task scheduler for distributed computations
    Copyright (C) 2012  Laurent Poirrier

    task is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <poll.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <limits.h>
#include <stdlib.h>
#include "block.h"
#include "date.h"
#include "array.h"
#include "buffer.h"
#include "config.h"
#include "network.h"
#include "queue.h"
#include "box.h"
#include "common.h"
#include "protocol.h"
#include "server.h"

// ***************************************************************************
// 
// ***************************************************************************
typedef struct server_t {
	int fd;
	boxset clients;
	queue done, pending;
	array pfd_a;
	
	int box_pool, task_pool;
} server;

#define DEBUG(s...)	\
	do { \
		fprintf(stderr, "%s ", curdate_str()); \
		fprintf(stderr, s); \
	} while(0)

// ***************************************************************************
// 
// ***************************************************************************
static void task_describe_header(buffer *msg, int flags)
{
	buffer_fappend(msg, "#%4s %23s %12s %1s %5s %26s",
		"task", "dir", "name", "S", "box", "host");
	
	if (flags & PROTO_TASK_DURATIONS)
		buffer_fappend(msg, "   %8s", "duration");
		
	if (flags & PROTO_TASK_TIMES)
		buffer_fappend(msg, "   %19s   %19s",
			"start time", "stop time");
	
	buffer_fappend(msg, "\n");
}

static void task_describe(buffer *msg, int flags,
	server *g, box *c, task *t, char *type)
{
	box *fc;
	int bid;
	char *host;
	char dstart[20], dstop[20];
	time_t dur;
	
	buffer_fappend(msg, "%5d %23s %12s %1s", t->id, t->dir, t->name, type);
	
	if (c) {
		bid = c->id;
		host = c->host;
	} else if (t->restrict_box >= 0) {
		bid = t->restrict_box;

		fc = boxset_find(&g->clients, t->restrict_box);
		if (fc)
			host = fc->host;
		else
			host = "-";
	} else {
		bid = -1;
		host = NULL;
	}
	
	if (bid >= 0)
		buffer_fappend(msg, " %5d %26s", bid, host);
	else
		buffer_fappend(msg, " %5s %26s", "", "");
	
	if (flags & PROTO_TASK_DURATIONS) {
		if (t->stop_time)
			dur = t->stop_time - t->start_time;
		else
			dur = time(NULL) - t->start_time;
		
		if (t->start_time)
			buffer_fappend(msg, "  %s%02d:%02d:%02d",
				(t->stop_time) ? " " : ">",
				dur / 3600, (dur / 60) % 60, dur % 60);
		else
			buffer_fappend(msg, "   %8s", "");
	}
	
	if (flags & PROTO_TASK_TIMES) {
		date_str(dstart, sizeof(dstart), t->start_time);
		date_str(dstop, sizeof(dstop), t->stop_time);
		
		if (!t->start_time)
			dstart[0] = 0;
		if (!t->stop_time)
			dstop[0] = 0;
		
		buffer_fappend(msg, "   %19s", dstart);
		buffer_fappend(msg, "   %19s", dstop);
	}

	buffer_fappend(msg, "\n");
}

// ***************************************************************************
// 
// ***************************************************************************
void box_describe_header(buffer *msg)
{
	buffer_fappend(msg, "#%4s %35s  R  C   in  out ref flags\n",
		"box", "host");
}

void box_describe(buffer *msg, box *fc)
{
	buffer_fappend(msg,
		"%5d %35s %2d %2d %4d %4d %3d %c%c\n",
		fc->id, fc->host, fc->running.size, fc->cores,
		fc->inbuff.size, fc->outbuff.size, fc->refs,
		(fc->flags & BOX_KICK) ? 'k' : '-',
		(fc->flags & BOX_SYNC) ? 'S' : '-');
}

// ***************************************************************************
// 
// ***************************************************************************
static void serv_fwd_unref(server *g, task *t)
{
	box *fc;
	
	if (t->forward_box < 0)
		return;
		
	fc = boxset_find(&g->clients, t->forward_box);
	
	if (fc) {
		fc->refs--;
	} else {
		DEBUG("[%d %s %s]: no fwd client (%d), EOF dismissed\n",
			t->id, t->dir, t->name, t->forward_box);
	}
}

// ***************************************************************************
// 
// ***************************************************************************
static int serv_sched_client(server *g, box *c)
{
	box *fc;
	task *t, *tnext;
	void *d;
	size_t dsz;
	char *s, *hs, *dir, *cmd, *name;
	int h, cores, id, flags, fd, rcode;
	buffer msg;
	
	if (c->outbuff.size > conf.proto_buff_max) {
		DEBUG("(%d %s): output buffer too large (%zd > %zd)\n",
			c->id, c->host, c->outbuff.size, conf.proto_buff_max);
		return(-1);
	}

	if (proto_prepare(&c->inbuff, PROTO_CL_MAGIC, conf.proto_buff_max)) {
		DEBUG("(%d %s): protocol error\n", c->id, c->host);
		return(-1);
	}
	
	if (!proto_ready(&c->inbuff))
		return(0);
	
	h = proto_header(&c->inbuff);

	// **********************************
	if (h == PROTO_CL_SETUP) {
		if (proto_fetch(&c->inbuff, "is", &cores, &s)) {
			DEBUG("(%d %s): PROTO_CL_SETUP: bad packet\n",
				c->id, c->host);
			return(-1);
		}
		
		DEBUG("(%d %s): setup as (%d %s) with %d cores\n",
			c->id, c->host, c->id, s, cores);
		
		box_setup(c, s, cores);
		return(0);
	}
		
	// **********************************
	if ((h == PROTO_CL_OUT) || (h == PROTO_CL_ERR)) {
		hs = (h == PROTO_CL_OUT) ? "OUT" : "ERR";
		
		if (proto_fetch(&c->inbuff, "id", &id, &d, &dsz)) {
			DEBUG("(%d %s): PROTO_CL_%s: bad packet\n",
				c->id, c->host, hs);
			return(-1);
		}

		t = queue_find(&c->running, id);
		if (!t) {
			DEBUG("(%d %s): PROTO_CL_%s: no task id %d\n",
				c->id, c->host, hs, id);
			return(-1);
		}
		
		// Forward data for task t from c to fc
		if (t->forward_box >= 0) {
			fc = boxset_find(&g->clients, t->forward_box);
			
			if (fc) {
				proto_append(&fc->outbuff, PROTO_SV_MAGIC,
					(h == PROTO_CL_OUT) ? PROTO_SV_OUT
						: PROTO_SV_ERR,
					"iisd", t->id, c->id, c->host, d, dsz);
			} else {
				DEBUG("[%d %s %s]: no fwd client (%d), "
					"data dismissed\n",
					t->id, t->dir, t->name, t->forward_box);
			}
		}
		
		// Write data for task t on server-side storage
		fd = (h == PROTO_CL_OUT) ? t->outfd : t->errfd;
		
		if (fd >= 0) {
			if (block_write(fd, d, dsz)) {
				fprintf(stderr, "%s: %s: %s\n", t->name,
					(h == PROTO_CL_OUT) ? t->outpath
						: t->errpath,
					strerror(errno));
				exit(1);
			}
		}
		
		return(0);
	}
	
	// **********************************
	if (h == PROTO_CL_END) {
		if (proto_fetch(&c->inbuff, "ii", &id, &rcode)) {
			DEBUG("(%d %s): PROTO_CL_END: bad packet\n",
				c->id, c->host);
			return(-1);
		}
		
		t = queue_find(&c->running, id);
		if (!t) {
			DEBUG("(%d %s): PROTO_CL_END: no task [%d]\n",
				c->id, c->host, id);
			return(-1);
		}

		DEBUG("[%d %s %s]: finished on (%d %s): %s\n",
			t->id, t->dir, t->name, c->id, c->host,
			(rcode == PROTO_END_DONE) ? "done"
			: ((rcode == PROTO_END_STOP) ? "stop" : "fail"));
			
		// Close server-local files:
		// if rcode is DONE, mark task as done
		// if rcode is STOP, delete task
		// if rcode is FAIL, remove files and requeue task

		if (rcode == PROTO_END_DONE) {
			task_close(t, 0);
			serv_fwd_unref(g, t);
			queue_unlink(&c->running, t);
			t->restrict_box = c->id;
			queue_append(&g->done, t);
		} else if (rcode == PROTO_END_STOP) {
			task_close(t, 0);
			serv_fwd_unref(g, t);
			queue_unlink(&c->running, t);
			task_free(t);
		} else if (rcode == PROTO_END_FAIL) {
			task_close(t, 1);
			queue_unlink(&c->running, t);
			queue_prepend(&g->pending, t);
		}
		
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_REQACK) {
		if (proto_fetch(&c->inbuff, "")) {
			DEBUG("(%d %s): PROTO_CL_REQACK: bad packet\n",
				c->id, c->host);
			return(-1);
		}
		
		buffer_init(&msg, conf.block_size);
		
		buffer_fappend(&msg, "%s: %zd\n",
			(c->ack_h == PROTO_CL_ADD) ? "Queued" :
			(c->ack_h == PROTO_CL_RM) ? "Removed" : "Matched",
			c->ack_n);
		
		c->ack_h = -1;
		c->ack_n = 0;
		
		proto_append(&c->outbuff, PROTO_SV_MAGIC, PROTO_SV_MSG,
			"d", msg.data, msg.size);
		c->refs = 0;

		buffer_clear(&msg);
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_ADD) {
		if (proto_fetch(&c->inbuff, "sss", &dir, &cmd, &name)) {
			DEBUG("(%d %s): PROTO_CL_ADD: bad packet\n",
				c->id, c->host);
			return(-1);
		}
		
		if (c->ack_h != h) {
			c->ack_h = h;
			c->ack_n = 0;
		}
		c->ack_n++;
		
		t = task_new(g->task_pool++, dir, cmd, name);
		queue_append(&g->pending, t);
		
		DEBUG("[%d %s %s]: added to queue\n",
			t->id, t->dir, t->name);

		return(0);
	}

	// **********************************
	if (h == PROTO_CL_EXEC) {
		if (proto_fetch(&c->inbuff, "s", &cmd)) {
			DEBUG("(%d %s): PROTO_CL_EXEC: bad packet\n",
				c->id, c->host);
			return(-1);
		}
		
		c->refs = 0;
		
		fc = g->clients.head;
		
		while (fc) {
			if (fc->cores >= 1) {
				t = task_new(g->task_pool++, "", cmd, "_exec_");
				t->restrict_box = fc->id;
				t->forward_box = c->id;
				c->refs++;
			
				queue_append(&g->pending, t);
				DEBUG("[%d  _exec_]: queued for sync "
					"on (%d %s)\n",
					t->id, fc->id, fc->host);
			}
			
			fc = fc->next;
		}

		return(0);
	}

	// **********************************
	if (h == PROTO_CL_RM) {
		if (proto_fetch(&c->inbuff, "iss", &flags, &dir, &name)) {
			DEBUG("(%d %s): PROTO_CL_STOP: bad packet\n",
				c->id, c->host);
			return(-1);
		}

		if (c->ack_h != h) {
			c->ack_h = h;
			c->ack_n = 0;
		}

		// For all matching done tasks, remove
		t = (flags & PROTO_TASK_DONE) ? g->done.head : NULL;
		while (t) {
			tnext = t->next;
			
			if (task_matches(t, dir, name)) {
				queue_unlink(&g->done, t);
				task_free(t);
				c->ack_n++;
			}
			
			t = tnext;
		}

		// For all matching running tasks, send stop cmd
		fc = (flags & PROTO_TASK_RUNNING) ? g->clients.head : NULL;
		while (fc) {
			t = fc->running.head;
			while (t) {
				if (task_matches(t, dir, name)) {
					proto_append(&fc->outbuff,
						PROTO_SV_MAGIC, PROTO_SV_STOP,
						"i", t->id);
					DEBUG("[%d %s %s]: interrupting"
						" on (%d %s)\n",
						t->id, t->dir, t->name,
						fc->id, fc->host);
					c->ack_n++;
				}
				t = t->next;
			}
			fc = fc->next;
		}
		
		// For all matching pending tasks, remove
		t = (flags & PROTO_TASK_PENDING) ? g->pending.head : NULL;
		while (t) {
			tnext = t->next;
		
			if (task_matches(t, dir, name)) {
				serv_fwd_unref(g, t);
				queue_unlink(&g->pending, t);
				task_free(t);
				c->ack_n++;
			}
		
			t = tnext;
		}
		
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_LIST) {
		if (proto_fetch(&c->inbuff, "i", &flags)) {
			DEBUG("(%d %s): PROTO_CL_LIST: bad packet\n",
				c->id, c->host);
			return(-1);
		}
		
		buffer_init(&msg, conf.block_size);
		task_describe_header(&msg, flags);
		
		if (flags & PROTO_TASK_DONE) {
			t = g->done.head;
			while (t) {
				task_describe(&msg, flags, g, NULL, t, ".");
				t = t->next;
			}
		}

		if (flags & PROTO_TASK_RUNNING) {
			fc = g->clients.head;
			while (fc) {
				t = fc->running.head;
				while (t) {
					task_describe(&msg, flags,
						g, fc, t, "r");
					t = t->next;
				}
				fc = fc->next;
			}
		}

		if (flags & PROTO_TASK_PENDING) {
			t = g->pending.head;
			while (t) {
				task_describe(&msg, flags, g, NULL, t, " ");
				t = t->next;
			}
		}

		proto_append(&c->outbuff, PROTO_SV_MAGIC, PROTO_SV_MSG,
			"d", msg.data, msg.size);
		c->refs = 0;
		
		buffer_clear(&msg);
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_NODES) {
		if (proto_fetch(&c->inbuff, "")) {
			DEBUG("(%d %s): PROTO_CL_NODES: bad packet\n",
				c->id, c->host);
			return(-1);
		}
		
		buffer_init(&msg, conf.block_size);
		box_describe_header(&msg);

		fc = g->clients.head;
		while (fc) {
			box_describe(&msg, fc);
			fc = fc->next;
		}

		proto_append(&c->outbuff, PROTO_SV_MAGIC, PROTO_SV_MSG,
			"d", msg.data, msg.size);
		c->refs = 0;
			
		buffer_clear(&msg);
		return(0);
	}

	// **********************************
	DEBUG("(%d %s): unknown request header 0x%08x\n",
		c->id, c->host, h);
	return(-1);
}

// ***************************************************************************
// 
// ***************************************************************************
static void serv_resched_running(server *g, box *c)
{
	task *t, *tnext;
	
	t = c->running.head;
	while (t) {
		tnext = t->next;
		
		// close, remove files, move to pending
		task_close(t, 1);
		queue_unlink(&c->running, t);
		queue_prepend(&g->pending, t);
		
		DEBUG("[%d %s %s]: requeueing\n", t->id, t->dir, t->name);
		
		t = tnext;
	}
}

static void serv_sched_go(server *g, box *c, task *t)
{
	queue_unlink(&g->pending, t);
	queue_append(&c->running, t);
	
	if (t->forward_box < 0)
		task_open(t);
	
	proto_append(&c->outbuff, PROTO_SV_MAGIC, PROTO_SV_RUN,
		"iss", t->id, conf.shell, t->cmd);
	
	DEBUG("[%d %s %s]: running on (%d %s)\n",
		t->id, t->dir, t->name, c->id, c->host);
}

// ***************************************************************************
// 
// ***************************************************************************
int serv_sched_task(server *g, task *t)
{
	box *c, *cbest;
	int rbest;
	
	// Restricted sync task
	if (t->restrict_box >= 0) {
		c = boxset_find(&g->clients, t->restrict_box);
		
		if (!c) {
			DEBUG("[%d %s %s]: restricted to missing box (%d)\n",
				t->id, t->dir, t->name, t->restrict_box);
			serv_fwd_unref(g, t);
			queue_unlink(&g->pending, t);
			task_free(t);
			return(0);
		}
		
		c->flags |= BOX_SYNC;
		if (c->running.size == 0)
			serv_sched_go(g, c, t);
		return(0);
	}
	
	cbest = NULL;
	rbest = INT_MAX;
	c = g->clients.head;
	while (c) {
		if ((!(c->flags & BOX_SYNC))
		&&  (c->running.size < c->cores)
		&&  (c->running.size < rbest)) {
			cbest = c;
			rbest = c->running.size;
		}
		c = c->next;
	}
	
	if (!cbest)
		return(-1);
	
	serv_sched_go(g, cbest, t);
	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
static int server_sched(server *g)
{
	box *c, *cnext;
	task *t, *tnext;
	
	// per-box handling
	c = g->clients.head;
	while (c) {
		cnext = c->next;
		
		if (serv_sched_client(g, c))
			c->flags |= BOX_KICK;
		
		if (c->running.size == 0)
			c->flags &= ~BOX_SYNC;
		
		// waste collection, part 1
		if (c->flags & BOX_KICK) {
			DEBUG("(%d %s): disconnect (had %d cores)\n",
				c->id, c->host, c->cores);
			serv_resched_running(g, c);
			close(c->fd);
			boxset_detach(&g->clients, c);
			box_free(c);
		}
		
		c = cnext;
	}
	
	// per-task handling
	t = g->pending.head;
	while (t) {
		tnext = t->next;

		if (serv_sched_task(g, t))
			break;

		t = tnext;
	}
	
	// waste collection, part 2
	c = g->clients.head;
	while (c) {
		cnext = c->next;

		if (buffer_io_complete(&c->outbuff)) {
			buffer_reset(&c->outbuff);
			buffer_io_reset(&c->outbuff);
			
			if (c->refs <= 0) {
				DEBUG("(%d %s): done forwarding "
					"(had %d cores)\n",
					c->id, c->host, c->cores);
				// Clients that are referenced should be
				// only capable of reading and displaying
				// data from server. Thus, they have no tasks.
				// So no 'serv_resched_running(g, c);'.
				// Btw, if they had, it would be a problem
				// as we would then need now to reschedule.
				close(c->fd);
				boxset_detach(&g->clients, c);
				box_free(c);
			}
		}
		
		c = cnext;
	}

	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
static int server_io(server *g)
{
	struct pollfd *pfd;
	netpeer cpeer;
	box *c;
	int i, cfd;
	
	// poll
	pfd = (struct pollfd *)array_get(&g->pfd_a, 1 + g->clients.n);
	
	pfd[0].fd = g->fd;
	pfd[0].events = POLLIN;
	pfd[0].revents = 0;
	
	c = g->clients.head;
	for (i = 0; i < g->clients.n; i++) {
		pfd[1 + i].fd = c->fd;
		pfd[1 + i].events = POLLIN;
		if (!buffer_io_complete(&c->outbuff))
			pfd[1 + i].events |= POLLOUT;
		pfd[1 + i].revents = 0;

		c = c->next;
	}

	if (poll(pfd, 1 + g->clients.n, -1) < 0) {
		fprintf(stderr, "server_iter(): poll(): %s\n", strerror(errno));
		return(-1);
	}
	
	// client i/o
	c = g->clients.head;
	for (i = 0; i < g->clients.n; i++) {
		if (pfd[1 + i].revents & POLLIN)
			buffer_io_read(&c->inbuff, c->fd);
		if (pfd[1 + i].revents & POLLOUT)
			buffer_io_write(&c->outbuff, c->fd);

		if ((c->inbuff.eof) || (c->inbuff.error)
		||  (c->outbuff.eof) || (c->outbuff.error)) {
			DEBUG("(%d %s): i/o terminated\n", c->id, c->host);
			c->flags |= BOX_KICK;
		}
		
		c = c->next;
	}
	
	// accept
	if (pfd[0].revents & POLLIN) {
		cfd = netserv_accept(g->fd, &cpeer);
		
		if (cfd >= 0) {
			c = box_new(g->box_pool++, cfd, &cpeer);
			boxset_attach(&g->clients, c);
			c->refs = 1;
			DEBUG("(%d %s): connected\n", c->id, c->host);
		} else {
			if ((errno != EINTR) && (errno != EAGAIN))
				return(-1);
		}
	}
	
	return(0);
}


// ***************************************************************************
// 
// ***************************************************************************
int task_server(const char *node, const char *service, int flags)
{
	server g;
	int r;
	
	g.fd = netserv_init(node, service, flags);
	
	if (g.fd < 0)
		return(1);

	boxset_init(&g.clients);
	queue_init(&g.done);
	queue_init(&g.pending);
	array_init(&g.pfd_a, sizeof(struct pollfd), 0);
	g.task_pool = 0;
	g.box_pool = 0;
	
	while (1) {
		if (server_io(&g)) {
			r = -1;
			break;
		}
		
		if (server_sched(&g)) {
			r = 0;
			break;
		}
	
	}
	
	boxset_clear(&g.clients);
	queue_clear(&g.done);
	queue_clear(&g.pending);
	array_clear(&g.pfd_a);
	
	close(g.fd);
	
	return(r);
}



